from setuptools import setup, find_packages
import os
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname), encoding='utf-8').read()
setup(
    name = "aRST-chem",
    version = "2.0.1",
    author = "Ying Chen",
    author_email = "ying.chen@pc.rwth-aachen.de",
    description = ("an Automated Reaction Search Tool"),
    license = "MIT",
    url = "https://git.rwth-aachen.de/bannwarthlab/aRST",
    packages=find_packages(),
    exclude_package_data={'': ['.gitignore']},
    include_package_data=True,
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.9',
    ],
    entry_points={
        'console_scripts': [
            'aRST = aRST.script.main:main',
        ]
    }
)
